/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.client.gui;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import javax.swing.SwingUtilities;

import kgrid.client.gui.View.EditingStatusType;
import kgrid.client.workflow.Edge;
import kgrid.client.workflow.Workflow;

import org.jgraph.graph.BasicMarqueeHandler;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.Port;
import org.jgraph.graph.PortView;

/**
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class MarqueeHandler extends BasicMarqueeHandler {

    protected Ports p;
    private Workflow workflow;
    private View clientView;

    public MarqueeHandler(View cv, Workflow w, Ports p) {
        clientView = cv;
        workflow = w;
        this.p = p;
    }

    private Object getCurrentCell(MouseEvent e) {
        return workflow.getFirstCellForLocation(e.getX(), e.getY());
    }

    // Override to Gain Control (for PopupMenu and ConnectMode)
    @Override
    public boolean isForceMarqueeEvent(MouseEvent e) {
        if (e.isShiftDown()) {
            return false;
        }
        // If Right Mouse Button we want to Display the PopupMenu
        if (SwingUtilities.isRightMouseButton(e)) // Return Immediately
        {
            return true;
        }
        // Find and Remember Port
        p.setPort(getSourcePortAt(e.getPoint()));
        // If Port Found and in ConnectMode (=Ports Visible)
        if (p.getPort() != null && workflow.isPortsVisible()) {
            return true;
        }
        // Else Call Superclass
        return super.isForceMarqueeEvent(e);
    }

    // Display PopupMenu or Remember Start Location and First Port
    @Override
    public void mousePressed(final MouseEvent e) {
        // If Right Mouse Button
        if (SwingUtilities.isRightMouseButton(e)) {
            // TODO remove the right click for the status switching
            // TODO implement the context menu with the right click
            // with two entries: "Connect with..." and "Delete"
            // with a confirmation dialog
            // TODO remove the edge and delete editing mode
            clientView.switchEditingStatus();
            workflow.setCursor(KGCursor
                    .getCursor(View.EditingStatusType.DRAG_PRESSED));
            e.consume();
            // Find Cell in Model Coordinates
            // Object cell = graph.getFirstCellForLocation(e.getX(), e.getY());
            // Create PopupMenu for the Cell
            // JPopupMenu menu = createPopupMenu(e.getPoint(), cell);
            // Display PopupMenu
            // menu.show(graph, e.getX(), e.getY());
            // Else if in ConnectMode and Remembered Port is Valid
        } else if (SwingUtilities.isLeftMouseButton(e)) {
            if (getCurrentCell(e) == null
                    && (View.getEditingStatus() == EditingStatusType.DATASET
                            || View.getEditingStatus() == EditingStatusType.TOOL || View
                            .getEditingStatus() == EditingStatusType.MODEL)) {
                workflow.addNode(e.getPoint());
                workflow.setSelectionCell(getCurrentCell(e));
            }
            // else it is taken care by the mouse handler
        } else if (p.getPort() != null && workflow.isPortsVisible()) {
            // Remember Start Location
            p.setStart(workflow.toScreen(p.getPort().getLocation()));
            // Remember First Port
            p.setFirstPort(p.getPort());
        } else {
            // Call Superclass
            super.mousePressed(e);
        }
    }

    // Find Port under Mouse and Repaint Connector
    public void mouseDragged(MouseEvent e) {
        // If remembered Start Point is Valid
        if (p.getStart() != null) {
            // Fetch Graphics from Graph
            Graphics g = workflow.getGraphics();
            // Reset Remembered Port
            PortView newPort = getTargetPortAt(e.getPoint());
            // Do not flicker (repaint only on real changes)
            if (newPort == null || newPort != p.getPort()) {
                // Xor-Paint the old Connector (Hide old Connector)
                paintConnector(Color.black, workflow.getBackground(), g);
                // If Port was found then Point to Port Location
                p.setPort(newPort);
                if (p.getPort() != null) {
                    p.setCurrent(workflow.toScreen(p.getPort().getLocation()));
                } // Else If no Port was found then Point to Mouse Location
                else {
                    p.setCurrent(workflow.snap(e.getPoint()));
                }
                // Xor-Paint the new Connector
                paintConnector(workflow.getBackground(), Color.black, g);
            }
        }
        // Call Superclass
        super.mouseDragged(e);
    }

    public PortView getSourcePortAt(Point2D point) {
        // Disable jumping
        workflow.setJumpToDefaultPort(false);
        PortView result;
        try {
            // Find a Port View in Model Coordinates and Remember
            result = workflow.getPortViewAt(point.getX(), point.getY());
        } finally {
            workflow.setJumpToDefaultPort(true);
        }
        return result;
    }

    // Find a Cell at point and Return its first Port as a PortView
    protected PortView getTargetPortAt(Point2D point) {
        // Find a Port View in Model Coordinates and Remember
        return workflow.getPortViewAt(point.getX(), point.getY());
    }

    // Connect the First Port and the Current Port in the Graph or Repaint
    @Override
    public void mouseReleased(MouseEvent e) {
        // If Valid Event, Current and First Port
        if (View.getEditingStatus() == View.EditingStatusType.EDGE
                || View.getEditingStatus() == View.EditingStatusType.DELETE
                || (SwingUtilities.isRightMouseButton(e) && View
                        .getEditingStatus() == View.EditingStatusType.DATASET)) {
            clientView.setNodeProperties(null);
        }
        if (e != null && p.getPort() != null && p.getFirstPort() != null
                && p.getFirstPort() != p.getPort()) {
            // Then Establish Connection
            workflow.connect((Port) p.getFirstPort().getCell(), (Port) p
                    .getPort().getCell(), null);
            e.consume();
            // Else Repaint the Graph
        } else {
            workflow.repaint();
        }
        // Reset Global Vars
        p.resetAll();
        // Call Superclass
        super.mouseReleased(e);
    }

    // Show Special Cursor if Over Port
    @Override
    public void mouseMoved(MouseEvent e) {
        // Check Mode and Find Port
        Object cell = getCurrentCell(e);
        if (e != null && cell != null && cell.getClass() != Edge.class
                && View.getEditingStatus() != View.EditingStatusType.EDGE
                && View.getEditingStatus() != View.EditingStatusType.DELETE) {
            workflow.setCursor(KGCursor
                    .getCursor(View.EditingStatusType.DRAG_HOVER));
            e.consume();
        } else if (e != null && getSourcePortAt(e.getPoint()) != null
                && workflow.isPortsVisible()) {
            // Set Cusor on Graph (Automatically Reset)
            workflow.setCursor(new Cursor(Cursor.HAND_CURSOR));
            // Consume Event
            // Note: This is to signal the BasicGraphUI's
            // MouseHandle to stop further event processing.

            e.consume();
        } else {
            super.mouseMoved(e);
        }
    }

    // Use Xor-Mode on Graphics to Paint Connector
    protected void paintConnector(Color fg, Color bg, Graphics g) {
        // Set Foreground
        g.setColor(fg);
        // Set Xor-Mode Color
        g.setXORMode(bg);
        // Highlight the Current Port
        paintPort(workflow.getGraphics());
        // If Valid First Port, Start and Current Point
        if (p.getFirstPort() != null && p.getStart() != null
                && p.getCurrent() != null) // Then Draw A Line From Start to
        // Current Point
        {
            g.drawLine((int) p.getStart().getX(), (int) p.getStart().getY(),
                    (int) p.getCurrent().getX(), (int) p.getCurrent().getY());
        }
    }

    // Use the Preview Flag to Draw a Highlighted Port
    protected void paintPort(Graphics g) {
        // If Current Port is Valid
        if (p.getPort() != null) {
            // If Not Floating Port...
            boolean o = (GraphConstants.getOffset(p.getPort()
                    .getAllAttributes()) != null);
            // ...Then use Parent's Bounds
            Rectangle2D r = (o) ? p.getPort().getBounds() : p.getPort()
                    .getParentView().getBounds();
            // Scale from Model to Screen
            r = workflow.toScreen((Rectangle2D) r.clone());
            // Add Space For the Highlight Border
            r.setFrame(r.getX() - 3, r.getY() - 3, r.getWidth() + 6, r
                    .getHeight() + 6);
            // Paint Port in Preview (=Highlight) Mode
            workflow.getUI().paintCell(g, p.getPort(), r, true);
        }
    }
}
