/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.client.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import javax.swing.AbstractAction;
import javax.swing.SwingUtilities;

import kgrid.client.gui.View.EditingStatusType;
import kgrid.client.workflow.DatasetNode;
import kgrid.client.workflow.Edge;
import kgrid.client.workflow.ToolNode;
import kgrid.client.workflow.Workflow;
import kgrid.client.workflow.WorkflowNode;

import org.jgraph.JGraph;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.PortView;

/**
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class MouseHandler implements MouseListener {

	protected Ports p;
	private Workflow workflow;
	private View view;

	public MouseHandler(View v, JGraph g, Ports p) {
		view = v;
		workflow = (Workflow) g;
		this.p = p;
	}

	protected AbstractAction remove = new AbstractAction("") {

		private static final long serialVersionUID = 1L;

		public void actionPerformed(ActionEvent e) {
			if (!workflow.isSelectionEmpty()) {
				workflow.removeSelectedCells();
			}
		}
	};

	public void showProperties(MouseEvent e) {
		if (SwingUtilities.isLeftMouseButton(e) && !workflow.isSelectionEmpty()) {
			// Object[] cells = graph.getSelectionCells();
			// cells = graph.getDescendants(cells);
			view.setNodeProperties((WorkflowNode) getCurrentCell(e));
		}
	}

	public void mouseClicked(MouseEvent e) {
	}

	// Use the Preview Flag to Draw a Highlighted Port
	protected void paintPort(Graphics g) {
		// If Current Port is Valid
		if (p.getPort() != null) {
			// If Not Floating Port...
			boolean o = (GraphConstants.getOffset(p.getPort()
					.getAllAttributes()) != null);
			// ...Then use Parent's Bounds
			Rectangle2D r = (o) ? p.getPort().getBounds() : p.getPort()
					.getParentView().getBounds();
			// Scale from Model to Screen
			r = workflow.toScreen((Rectangle2D) r.clone());
			// Add Space For the Highlight Border
			r.setFrame(r.getX() - 3, r.getY() - 3, r.getWidth() + 6,
					r.getHeight() + 6);
			// Paint Port in Preview (=Highlight) Mode
			workflow.getUI().paintCell(g, p.getPort(), r, true);
		}
	}

	// Use Xor-Mode on Graphics to Paint Connector
	protected void paintConnector(Color fg, Color bg, Graphics g) {
		// Set Foreground
		g.setColor(fg);
		// Set Xor-Mode Color
		g.setXORMode(bg);
		// Highlight the Current Port
		paintPort(workflow.getGraphics());
		// If Valid First Port, Start and Current Point
		if (p.getFirstPort() != null && p.getStart() != null
				&& p.getCurrent() != null) // Then Draw A Line From Start to
		// Current Point
		{
			g.drawLine((int) p.getStart().getX(), (int) p.getStart().getY(),
					(int) p.getCurrent().getX(), (int) p.getCurrent().getY());
		}
	}

	// Find a Cell at point and Return its first Port as a PortView
	protected PortView getTargetPortAt(Point2D point) {
		// Find a Port View in Model Coordinates and Remember
		return workflow.getPortViewAt(point.getX(), point.getY());
	}

	public void mouseDragged(MouseEvent e) {
		// If remembered Start Point is Valid
		if (p.getStart() != null) {
			// Fetch Graphics from Graph
			Graphics g = workflow.getGraphics();
			// Reset Remembered Port
			PortView newPort = getTargetPortAt(e.getPoint());
			// Do not flicker (repaint only on real changes)
			if (newPort == null || newPort != p.getPort()) {
				// Xor-Paint the old Connector (Hide old Connector)
				paintConnector(Color.black, workflow.getBackground(), g);
				// If Port was found then Point to Port Location
				p.setPort(newPort);
				if (p.getPort() != null) {
					p.setCurrent(workflow.toScreen(p.getPort().getLocation()));
				} // Else If no Port was found then Point to Mouse Location
				else {
					p.setCurrent(workflow.snap(e.getPoint()));
				}
				// Xor-Paint the new Connector
				paintConnector(workflow.getBackground(), Color.black, g);
			}
		}
		// Call Superclass
		// super.mouseDragged(e);
	}

	/**
	 * Responsible for dealing with edges.
	 */
	public void mousePressed(MouseEvent e) {
		Object cell = getCurrentCell(e);
		if (e != null && cell != null && SwingUtilities.isLeftMouseButton(e)) {
			if (cell.getClass() != Edge.class) {
				switch (View.getEditingStatus()) {
				case EDGE:
					p.setPort(getSourcePortAt(e.getPoint()));
					if (p.getPort() != null) {
						p.setStart(workflow.toScreen(p.getPort().getLocation()));
						p.setFirstPort(p.getPort());
					}
					break;
				case DELETE:
					remove.actionPerformed(null);
					break;

				default:
					showProperties(e);
					workflow.setCursor(KGCursor
							.getCursor(View.EditingStatusType.DRAG_PRESSED));
					WorkflowNode node = (WorkflowNode) getCurrentCell(e);
					String kdsURL = node.getKDSURL();
					if (kdsURL != null) {
						view.setStatus(kdsURL);
					} else {
						view.setStatus("");
					}
					e.consume();
					break;
				}
			} else if (View.getEditingStatus() == EditingStatusType.DELETE) {
				remove.actionPerformed(null);
				e.consume();
			} else if (cell instanceof Edge) {
				// Disable search button
				view.setSearchButtonEnabled(false);
				Edge edge = (Edge) cell;
				if (edge.isLabelEditable()) {
					view.setEdgeProperties(edge);
					workflow.setSelectionCell(getCurrentCell(e));
					e.consume();
				} else {
					view.setNodeProperties(null);
				}
			}

		} else if (SwingUtilities.isRightMouseButton(e)) {
			workflow.setCursor(KGCursor.getCursor());
			e.consume();
		} else {
			e.consume();
		}
	}

	public PortView getSourcePortAt(Point2D point) {
		// Disable jumping
		workflow.setJumpToDefaultPort(false);
		PortView result;
		try {
			// Find a Port View in Model Coordinates and Remember
			result = workflow.getPortViewAt(point.getX(), point.getY());
		} finally {
			workflow.setJumpToDefaultPort(true);
		}
		return result;
	}

	/**
	 * Responsible for dealing with nodes.
	 */
	public void mouseReleased(MouseEvent e) {
		Object cell = getCurrentCell(e);
		if (SwingUtilities.isRightMouseButton(e)
				|| (View.getEditingStatus() == View.EditingStatusType.EDGE && (e == null
						|| cell == null || !(cell instanceof Edge)))
				|| View.getEditingStatus() == View.EditingStatusType.DELETE) {
			view.setNodeProperties(null);
		}
		if (e != null && cell != null && cell.getClass() != Edge.class
				&& SwingUtilities.isLeftMouseButton(e)) {
			// Check for enabling/disabling the search button
			if ((cell instanceof DatasetNode || cell instanceof ToolNode)
					&& view.getPropertiesPanel().getType()
							.equals(PropertiesPanel.NODE_PANEL)) {
				view.setSearchButtonEnabled(true);
			} else {
				view.setSearchButtonEnabled(false);
			}
			if (View.getEditingStatus() != View.EditingStatusType.EDGE
					&& View.getEditingStatus() != View.EditingStatusType.DELETE) {
				workflow.setCursor(KGCursor
						.getCursor(View.EditingStatusType.DRAG_HOVER));
			}
			e.consume();
		}
	}

	private Object getCurrentCell(MouseEvent e) {
		return workflow.getFirstCellForLocation(e.getX(), e.getY());
	}

	public void mouseEntered(MouseEvent e) {
		workflow.setCursor(KGCursor.getCursor());
		e.consume();
	}

	public void mouseExited(MouseEvent e) {
	}

}
