/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.client.workflow;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Map;

import org.jdesktop.application.ResourceMap;
import org.jdom.Element;
import org.jdom.JDOMException;

/**
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class DatasetNode extends WorkflowNode {

	private static final long serialVersionUID = 1L;
	private static final String NAME = "Dataset";
	private static final int TYPE = DATASET_TYPE;

	private String matchedKDSURL = null;

	private LinkedList<LinkedHashMap<String, Object>> attributes = new LinkedList<LinkedHashMap<String, Object>>();

	public DatasetNode(int id, int x, int y, ResourceMap resourceMap, Workflow w) {
		super(id, NAME, x, y, resourceMap, w);
		properties.put(DatasetNode.PKDS, "");
		properties.put(PNAME, "");
		properties.put(PTYPE, "");
		properties.put(PSIZE, ""); // number of instances
		properties.put(PASEP, "");
		properties.put(PRSEP, "");
		properties.put(PUTOK, "");
		properties.put(PATTR, attributes);
	}

	public boolean isEmpty() {
		return ((String) properties.get(PKDS)).isEmpty()
				&& ((String) properties.get(PNAME)).isEmpty()
				&& ((String) properties.get(PTYPE)).isEmpty()
				&& ((String) properties.get(PSIZE)).isEmpty()
				&& ((String) properties.get(PASEP)).isEmpty()
				&& ((String) properties.get(PRSEP)).isEmpty()
				&& ((String) properties.get(PUTOK)).isEmpty()
				&& attributes.isEmpty();
	}

	@Override
	public void clearSearchParameters() {
		properties.put(PNAME, "");
		properties.put(PTYPE, "");
		properties.put(PSIZE, "");
		properties.put(PASEP, "");
		properties.put(PRSEP, "");
		properties.put(PUTOK, "");
		properties.put(PATTR, attributes);
		attributes.clear();
	}

	@Override
	public String getKDSURL() {
		return matchedKDSURL;
	}

	@Override
	protected void setKDSURL(String s) {
		matchedKDSURL = s;
	}

	public Object getProperty(String key) {
		if (!key.equals(PATTR)) {
			properties.get(key);
		}
		return null;
	}

	public LinkedList<LinkedHashMap<String, Object>> getAttrList() {
		return attributes;
	}

	public void setProperties(Object description) throws JDOMException {
		if (description != null) {
			Element dataset = (Element) description;
			if (dataset.getName().equals("Dataset")) {
				String href = dataset.getAttributeValue("href");
				if (href != null && !href.isEmpty()
						&& !href.equals(KDSURL_PREFIX)) {
					properties.put(DatasetNode.PKDS, href);
					matchedKDSURL = href;
				} else {
					String name = dataset.getAttributeValue("name");
					if (name != null) {
						properties.put(PNAME, name);
					}
					String type = dataset.getAttributeValue("type");
					if (type != null) {
						properties.put(PTYPE, type);
					}
				}
				for (Object datasetChild : dataset.getChildren()) {
					Element dataElement = (Element) datasetChild;
					if (dataElement.getName().equals("Access")) {
						for (Object accessChild : dataElement.getChildren()) {
							Element accessElement = (Element) accessChild;
							if (accessElement.getName().equals("Size")) {
								properties.put(PSIZE, accessElement.getText());
							} else {
								throw new JDOMException("Tag <"
										+ accessElement.getName() + "> unknown");
							}
						}
					} else if (dataElement.getName().equals("Structure")) {
						for (Object structureChild : dataElement.getChildren()) {
							Element structureElement = (Element) structureChild;
							if (structureElement.getName().equals("Format")) {
								for (Object formatChild : structureElement
										.getChildren()) {
									Element formatElement = (Element) formatChild;
									properties.put(formatElement.getName(),
											formatElement.getText());
								}
							} else if (structureElement.getName().equals(
									"Attributes")) {
								addAttributes(structureElement);
							} else {
								throw new JDOMException("Tag <"
										+ structureElement.getName()
										+ "> unknown");
							}
						}
					} else {
						throw new JDOMException("Tag <" + dataElement.getName()
								+ "> unknown");
					}
				}
			} else {
				throw new JDOMException("Tag <" + dataset.getName()
						+ "> unknown");
			}
			resetStatus();
		}
	}

	private void addAttributes(Element attributes) throws JDOMException {
		for (Object attributesChild : attributes.getChildren()) {
			Element attribute = (Element) attributesChild;
			if (!attribute.getName().equals("Attribute")) {
				throw new JDOMException("Tag <" + attribute.getName()
						+ "> unknown");
			}

			LinkedHashMap<String, Object> attrProperties = new LinkedHashMap<String, Object>();

			attrProperties.put(ATTRNAME, attribute.getAttributeValue("name"));
			attrProperties.put(PTYPE, attribute.getAttributeValue("type"));
			LinkedList<String> parameters = new LinkedList<String>();
			for (Object attributeChild : attribute.getChildren()) {
				Element attributeElement = (Element) attributeChild;
				if (attributeElement.getName().equals("Parameter")) {
					parameters.add(attributeElement.getText());
				} else {
					attrProperties.put(attributeElement.getName(),
							attributeElement.getText());
				}
			}
			attrProperties.put(PARAM, parameters);

			this.attributes.add(attrProperties);

		}
	}

	public void addAttribute() {
		LinkedHashMap<String, Object> attrProperties = new LinkedHashMap<String, Object>();
		attrProperties.put(ATTRNAME, "");
		attrProperties.put(PTYPE, "");
		attrProperties.put(PSTYPE, "");
		LinkedList<String> parameters = new LinkedList<String>();
		parameters.add("");
		parameters.add("");
		attrProperties.put(PARAM, parameters);
		this.attributes.add(attrProperties);
	}

	@Override
	public String toString() {
		String kdsUrl = (String) properties.get(PKDS);
		if (!kdsUrl.isEmpty() && !kdsUrl.equals(KDSURL_PREFIX)) {
			kdsUrl = kdsUrl.substring(kdsUrl.lastIndexOf("/") + 1);
			int ext = kdsUrl.lastIndexOf(".");
			if (ext > 0) {
				return kdsUrl.substring(0, ext);
			}
			return kdsUrl;
		}
		return (String) properties.get(PNAME);
	}

	@Override
	public int getType() {
		return TYPE;
	}

	/**
	 * Sets the value of a property of the dataset node.
	 * 
	 * @param row
	 *            the index of the row it has been edited in the properties
	 *            panel
	 * @param property
	 *            the name of the property to be set
	 * @param value
	 *            the value of the property to be set
	 */
	@Override
	public void setProperty(int row, String property, String value) {
		if (row >= properties.size() - 1) {
			/*
			 * it's an attribute property: now we have to find out first which
			 * attribute is to be set
			 */
			int iAttr = 0;
			row -= properties.size() - 1;
			ListIterator<LinkedHashMap<String, Object>> it = attributes
					.listIterator();
			while (it.hasNext()) {
				LinkedHashMap<String, Object> attribute = it.next();
				LinkedList<String> parameters = (LinkedList<String>) attribute
						.get(PARAM);
				if (row < attribute.size() + parameters.size() - 1) {
					/*
					 * we have found the attribute, now we look for the
					 * parameter
					 */
					if (row < attribute.size() - 1) {
						// primary parameter
						switch (row) {
						case 0:
							attribute.put(ATTRNAME, value);
							break;
						case 1:
							attribute.put(PTYPE, value);
							break;
						case 2:
							attribute.put(PSTYPE, value);
							break;

						default:
							System.err.println("row (" + row
									+ ") index out of limits");
						}

					} else {
						// secondary parameter
						row -= (attribute.size() - 1);
						parameters.set(row, value);
					}
					return;
				}
				row -= attribute.size() + parameters.size() - 1;
				iAttr++;
			}
		} else {
			// it's a dataset property
			properties.put(property, value);
		}
		resetStatus();
	}

	@Override
	public GraphElement getElement() {
		GraphElement nodeE = new GraphElement("node", getIdName());
		nodeE.setAttribute("id", getIdName());
		Element typeE = new Element("data");
		typeE.setAttribute("key", "type");
		typeE.addContent("dataset");
		nodeE.addContent(typeE);
		if (!isEmpty()) {
			Element descriptionE = new Element("data");
			descriptionE.setAttribute("key", "description");
			nodeE.addContent(descriptionE);

			Element datasetE = new Element("Dataset");

			String name = (String) properties.get(PNAME);
			if (!name.isEmpty()) {
				datasetE.setAttribute("name", name);
			} else {
				// TODO to be removed when the EPMS will be updated
				datasetE.setAttribute("name", "");
			}

			String type = (String) properties.get(PTYPE);
			if (!type.isEmpty()) {
				datasetE.setAttribute("type", type);
			}

			String kdsUrl = (String) properties.get(DatasetNode.PKDS);
			if (!kdsUrl.isEmpty() && !kdsUrl.equals(KDSURL_PREFIX)) {
				datasetE.setAttribute("href", kdsUrl);
			}

			String size = (String) properties.get(PSIZE);
			if (!size.isEmpty()) {
				Element accessE = new Element("Access");
				Element sizeE = new Element("Size");
				sizeE.setText(size);
				accessE.addContent(sizeE);
				datasetE.addContent(accessE);
			}

			String attSep = (String) properties.put(PASEP, "");
			String recSep = (String) properties.put(PRSEP, "");
			String uTok = (String) properties.put(PUTOK, "");

			boolean isFormatEmpty = attSep.isEmpty() && recSep.isEmpty()
					&& uTok.isEmpty();

			boolean isAttributesEmpty = attributes.isEmpty();

			if (!isFormatEmpty || !isAttributesEmpty) {
				Element structureE = new Element("Structure");
				if (!isFormatEmpty) {
					Element formatE = new Element("Format");
					if (!attSep.isEmpty()) {
						Element attSepE = new Element(PASEP);
						attSepE.setText(attSep);
						formatE.addContent(attSepE);
					}

					if (!recSep.isEmpty()) {
						Element recSepE = new Element(PRSEP);
						recSepE.setText(recSep);
						formatE.addContent(recSepE);
					}
					if (!uTok.isEmpty()) {
						Element uTokE = new Element(PUTOK);
						uTokE.setText(uTok);
						formatE.addContent(uTokE);
					}
					structureE.addContent(formatE);
				}
				if (!isAttributesEmpty) {
					Element attributesE = new Element("Attributes");
					for (LinkedHashMap<String, Object> attr : attributes) {
						Element attrE = new Element("Attribute");

						String attrName = (String) attr.get(ATTRNAME);
						if (attrName != null && !attrName.isEmpty()) {
							attrE.setAttribute("name", attrName);
						}

						String attrType = (String) attr.get(PTYPE);
						if (attrType != null && !attrType.isEmpty()) {
							attrE.setAttribute("type", attrType);
						}
						for (Map.Entry<String, Object> e : attr.entrySet()) {
							if (e.getValue() instanceof LinkedList<?>) {
								LinkedList<String> parameters = (LinkedList<String>) e
										.getValue();
								for (String p : parameters) {
									Element elem = new Element("Parameter");
									elem.setText(p);
									attrE.addContent(elem);
								}
							} else {
								String key = e.getKey();
								if (!key.equals(ATTRNAME) && !key.equals(PTYPE)) {
									Element elem = new Element(key);
									elem.setText((String) e.getValue());
									attrE.addContent(elem);
								}
							}
						}
						attributesE.addContent(attrE);
					}
					structureE.addContent(attributesE);
				}
				datasetE.addContent(structureE);
			}
			descriptionE.addContent(datasetE);
		}
		addPosition(nodeE);
		return nodeE;
	}

	@Override
	public void resetStatus() {
		if (isEmpty()) {
			setStatus(NOINFO);
		} else {
			String kdsUrl = (String) properties.get(PKDS);
			if (kdsUrl.isEmpty() || kdsUrl.equals(KDSURL_PREFIX)) {
				setStatus(SEARCHP_SET);
			} else {
				setStatus(KDSURL_SET);
			}
		}
		if (((String) properties.get(PKDS)).isEmpty()) {
			matchedKDSURL = null;
		}
	}
}
