/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.client.workflow;

import java.awt.Color;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import javax.swing.JOptionPane;

import kgrid.client.gui.MarqueeHandler;
import kgrid.client.gui.MouseHandler;
import kgrid.client.gui.Ports;
import kgrid.client.gui.Settings;
import kgrid.client.gui.Utils;
import kgrid.client.gui.View;
import kgrid.client.gui.WheelHandler;
import kgrid.client.remote.EPMSTask;

import org.apache.commons.lang3.time.DurationFormatUtils;
import org.apache.commons.lang3.time.StopWatch;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.jgraph.JGraph;
import org.jgraph.event.GraphModelEvent;
import org.jgraph.event.GraphModelListener;
import org.jgraph.graph.BasicMarqueeHandler;
import org.jgraph.graph.DefaultCellViewFactory;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.DefaultPort;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.GraphLayoutCache;
import org.jgraph.graph.GraphModel;
import org.jgraph.graph.Port;

/**
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class Workflow extends JGraph implements Consts {

	private static final long serialVersionUID = 1L;

	private boolean changed = false;
	private boolean running = false;
	private StopWatch stopWatch = new StopWatch();

	private View view;
	private String fileFullPath = DEFAULT_FILE_NAME;
	private BasicMarqueeHandler basicMarqueeHandler = new BasicMarqueeHandler();

	private MarqueeHandler customMarqueeHandler;

	public Workflow(View view) {

		// System.out.println(JGraph.VERSION);
		this.view = view;
		GraphModel model = new DefaultGraphModel();
		GraphLayoutCache cache = new GraphLayoutCache(model,
				new DefaultCellViewFactory());

		setModel(model);
		setGraphLayoutCache(cache);

		setAntiAliased(true);
		setSizeable(false);
		setDisconnectable(false);
		setOpaque(true);
		setBendable(false);
		setHandleSize(0);
		setEditable(false);
		setPortsOnTop(false);
		setGridSize(20);

		setToolTipText("");

		setBackground(Color.white);
		setLockedHandleColor(Color.blue);
		setHighlightColor(Color.blue);

		Ports ports = new Ports();
		customMarqueeHandler = new MarqueeHandler(view, this, ports);
		setMarqueeHandler(customMarqueeHandler);
		addMouseListener(new MouseHandler(view, this, ports));
		addMouseWheelListener(new WheelHandler(view));

		getModel().addGraphModelListener(new ModelListener());

		getGraphLayoutCache().setAutoSizeOnValueChange(true);

	}

	public void start() {
		stopWatch.reset();
		stopWatch.start();
		Utils.print("The workflow \"" + getFileName() + "\" has started");
		running = true;
		EPMSTask task = null;
		try {
			task = new EPMSTask(this);
			task.invoke();
		} catch (Exception e) {
			String exMessage = e.toString();
			int endPos = exMessage.indexOf(":");
			String message = exMessage.substring(endPos + 2);
			if (message.contains("Connection timed out")) {
				message += "\nThe connection may be blocked\nby the client or the EPMS host firewall.";
			} else if (message.contains("Connection refused")) {
				message += "\nThe Globus container at "
						+ Settings.getEPMSAddress() + ":"
						+ Settings.getEPMSPort()
						+ " may not have been started.";
			} else if (message.contains("No route to host ")) {
				message += Settings.getEPMSAddress() + ":"
						+ Settings.getEPMSPort()
						+ " is unreachable (it may be down).";
			}
			String exMessageClassName = exMessage;
			int startPos = exMessage.substring(0, endPos).lastIndexOf(".");
			if (0 <= startPos && startPos <= endPos) {
				exMessageClassName = exMessage.substring(startPos + 1, endPos);
			}
			if (task != null) {
				kgrid.client.remote.Utils.print(exMessage, task.getEPR()
						.getAddress());
			} else {
				Utils.print(exMessage);
			}
			// remove this printStackTrace when not debugging
			// e.printStackTrace();
			JOptionPane.showMessageDialog(view.getFrame(), message,
					exMessageClassName, JOptionPane.ERROR_MESSAGE);
		}
	}

	@Override
	public String getToolTipText(MouseEvent e) {
		String tip = null;
		Object cell = getFirstCellForLocation(e.getX(), e.getY());
		if (e != null && cell != null && cell instanceof WorkflowNode) {
			tip = ((WorkflowNode) cell).getStatusString();
		}
		return tip;
	}

	public void clear() {
		fileFullPath = DEFAULT_FILE_NAME;
		getModel().remove(getDescendants(getRoots()));
		getModel().remove(getRoots());
		clearSelection();
	}

	public Object getSelectedCell() {
		return getSelectionCell();
	}

	public boolean isFileNameSet() {
		return !fileFullPath.equals(DEFAULT_FILE_NAME);
	}

	public boolean hasChanged() {
		return changed;
	}

	public void setChanged(boolean c) {
		changed = c;
		if (changed) {

			getGraphLayoutCache().showCells(getRoots(), true);

			view.setTabTitle("*"
					+ fileFullPath.substring(fileFullPath
							.lastIndexOf(File.separator) + 1));
			view.enableFileMenuItems();
		} else {
			view.setTabTitle(fileFullPath.substring(fileFullPath
					.lastIndexOf(File.separator) + 1));
		}
	}

	class ModelListener implements GraphModelListener {
		@Override
		public void graphChanged(GraphModelEvent arg0) {
			if (!changed) {
				view.enableFileMenuItems();
				changed = true;
				if (fileFullPath.equals(DEFAULT_FILE_NAME)) {
					view.setTabTitle("*" + DEFAULT_FILE_NAME);
				} else {
					view.setTabTitle("*"
							+ fileFullPath.substring(fileFullPath
									.lastIndexOf(File.separator) + 1));
				}

			}
		}
	}

	public void loadFromFile(String newFileName) {
		fileFullPath = newFileName;
		SAXBuilder builder = new SAXBuilder();
		try {
			Document doc = builder.build(fileFullPath);
			Element graphml = doc.getRootElement();
			if (graphml.getName().equals("graphml")) {
				Element graph = (Element) graphml.getChildren().get(0);
				if (graph.getName().equals("graph")) {
					List<?> nodes = graph.getChildren();
					for (Object child : nodes) {
						Element e = (Element) child;
						String eName = e.getName();
						if (eName.equals("node")) {
							addNodeFromXML(e);
						} else if (eName.equals("edge")) {
							// TODO bug fix: sometimes label appears far from
							// its edge when opening a workflow with nodes with
							// not fixed positions (when two nodes are very
							// close to each other)
							addEdge(e);
						} else {
							throw new JDOMException("Tag <" + eName
									+ "> unknown");
						}
					}
				} else {
					throw new JDOMException("Tag <" + graph.getName()
							+ "> unknown");
				}
			} else {
				throw new JDOMException("Tag <" + graphml.getName()
						+ "> unknown");
			}
		} catch (JDOMException e) {
			JOptionPane.showMessageDialog(view.getFrame(), e, "JDOMException",
					JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
		} catch (IOException e) {
			JOptionPane.showMessageDialog(view.getFrame(), e, "IOException",
					JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
		}
	}

	private void addNodeFromXML(Element node) throws JDOMException {
		List<?> children = node.getChildren();
		String id = node.getAttributeValue("id");
		String type = null;
		String location = null;
		String resultDate = null;
		String resultFileName = null;
		Element description = null;
		int posX = getNumberOfNodes() * 40 + 40;
		int posY = ((int) (Math.random() * 10)) * 40 + 40;
		for (Object child : children) {
			Element data = (Element) child;
			String value = data.getAttributeValue("key");
			if (value == null) {
				throw new JDOMException("Attribute key not present in data tag");
			}
			if (value.equals("type")) {
				type = data.getTextNormalize();
			} else if (value.equals("status")) {
				// status value ignored
			} else if (value.equals("description") || value.equals("metadata")) {
				// TODO xml bug: description or metadata tag? IGNORE metadata
				if (!data.getChildren().isEmpty()) {
					description = (Element) data.getChildren().get(0);
				}
			} else if (value.equals("location")) {
				location = data.getTextNormalize();
			} else if (value.equals("result")) {
				resultDate = data.getAttributeValue("date");
				resultFileName = data.getTextNormalize();
			} else if (value.equals("position_X")) {
				posX = Integer.parseInt(data.getTextNormalize());
			} else if (value.equals("position_Y")) {
				posY = Integer.parseInt(data.getTextNormalize());
			} else {
				throw new JDOMException("Tag <" + value + "> unknown");
			}
		}
		WorkflowNode wNode;
		if (type.equals("dataset")) {
			wNode = getDatasetNode(posX, posY);
		} else if (type.equals("algorithm")) {
			wNode = getToolNode(posX, posY);
		} else if (type.equals("model")) {
			wNode = getModelNode(posX, posY);
		} else {
			throw new JDOMException("Tag <" + type + "> unknown");
		}
		wNode.setProperties(description);
		if (location != null) {
			if (wNode instanceof ModelNode) {
				((ModelNode) wNode)
						.setProperty(0, ModelNode.LOCATION, location);
			} else {
				// TODO xml bug: dataset nodes in Modello4.xml have the location
				// tag ===> IGNORE IT
			}
		}
		if (resultDate != null) {
			if (wNode instanceof ModelNode) {
				((ModelNode) wNode).addResult(resultDate, resultFileName);
			} else {
			}
		}
		wNode.setId(id);
		getGraphLayoutCache().insert(wNode);
	}

	private int getNumberOfNodes() {
		int n = 0;
		for (Object cell : getRoots()) {
			if (cell instanceof WorkflowNode) {
				n++;
			}
		}
		return n;
	}

	public ArrayList<Edge> getOutgoingEdges(WorkflowNode node) {
		ArrayList<Edge> edges = new ArrayList<Edge>();
		for (Object cell : getRoots()) {
			if (cell instanceof Edge) {
				if (((Edge) cell).getSourceNode().equals(node)) {
					edges.add((Edge) cell);
				}
			}
		}
		return edges;
	}

	private Edge getFirstOutgoingEdge(WorkflowNode node) {
		for (Object cell : getRoots()) {
			if (cell instanceof Edge) {
				if (((Edge) cell).getSourceNode().equals(node)) {
					return (Edge) cell;
				}
			}
		}
		return null;
	}

	// Insert a new Vertex at point
	public void addNode(Point2D point) {
		switch (View.getEditingStatus()) {
		case DATASET:
			addDatasetNode((int) point.getX(), (int) point.getY());
			break;
		case TOOL:
			addToolNode((int) point.getX(), (int) point.getY());
			break;
		case MODEL:
			addModelNode((int) point.getX(), (int) point.getY());
			break;
		default:
			JOptionPane.showMessageDialog(view.getFrame(), "Editing status <"
					+ View.getEditingStatus() + "> unknown",
					"Error trying to add a node", JOptionPane.ERROR_MESSAGE);
			break;
		}
	}

	/*
	 * Returns an LinkedList containing the workflow nodes objects
	 * 
	 * @return a list containing the workflow nodes objects
	 */
	private LinkedList<WorkflowNode> getNodes() {
		LinkedList<WorkflowNode> nodes = new LinkedList<WorkflowNode>();
		for (Object cell : getRoots()) {
			if (cell instanceof WorkflowNode) {
				nodes.add((WorkflowNode) cell);
			}
		}
		return nodes;
	}

	/**
	 * Returns an LinkedList containing the workflow edges
	 * 
	 * @return a list containing the workflow edges
	 */
	public LinkedList<Edge> getEdges() {
		LinkedList<Edge> edges = new LinkedList<Edge>();
		for (Object cell : getRoots()) {
			if (cell instanceof Edge) {
				edges.add((Edge) cell);
			}
		}
		return edges;
	}

	/*
	 * Searches an unused id, starting from 0, among the ids of the current
	 * workflow nodes
	 * 
	 * @return a unused id among the current workflow nodes
	 */
	private int generateNewNodeID() {
		int id = 0;
		boolean idExists = true;
		LinkedList<WorkflowNode> nodes = getNodes();
		while (idExists) {
			idExists = false;
			for (WorkflowNode node : nodes) {
				if (node.getIdNumber() == id) {
					idExists = true;
					id++;
					break;
				}
			}
		}
		return id;
	}

	/*
	 * Searches an unused id, starting from 0, among the ids of the current
	 * workflow edges
	 * 
	 * @return a unused id among the current workflow edges
	 */
	private int generateNewEdgeID() {
		int id = 0;
		boolean idExists = true;
		LinkedList<Edge> edges = getEdges();
		while (idExists) {
			idExists = false;
			for (Edge edge : edges) {
				if (edge.getId() == id) {
					idExists = true;
					id++;
					break;
				}
			}
		}
		return id;
	}

	private WorkflowNode getDatasetNode(int x, int y) {
		return new DatasetNode(generateNewNodeID(), x, y,
				view.getResourceMap(), this);
	}

	private WorkflowNode getToolNode(int x, int y) {
		return new ToolNode(generateNewNodeID(), x, y, view.getResourceMap(),
				this);
	}

	private WorkflowNode getModelNode(int x, int y) {
		return new ModelNode(generateNewNodeID(), x, y, view.getResourceMap(),
				this);
	}

	private void addDatasetNode(int x, int y) {
		getGraphLayoutCache().insert(getDatasetNode(x, y));
	}

	private void addToolNode(int x, int y) {
		getGraphLayoutCache().insert(getToolNode(x, y));
	}

	private void addModelNode(int x, int y) {
		getGraphLayoutCache().insert(getModelNode(x, y));
	}

	private void addEdge(Element edge) throws JDOMException {
		String sourceId = edge.getAttributeValue("source");
		String targetId = edge.getAttributeValue("target");
		String type = null;
		for (Object edgeChild : edge.getChildren()) {
			Element edgeElement = (Element) edgeChild;
			if (edgeElement.getName().equals("data")) {
				if (edgeElement.getAttributeValue("key").equals("type")) {
					type = edgeElement.getText();
					if (!type.equals("dataset") && !type.equals("train")
							&& !type.equals("test") && !type.equals("model")
							&& !type.equals("transfer") && !type.equals("?")) {
						throw new JDOMException("Edge type \"" + type
								+ "\" unknown");
					}
				} else {
					throw new JDOMException("Key attribute value \""
							+ edgeElement.getAttributeValue("key")
							+ "\" unknown");
				}
			} else {
				throw new JDOMException("Tag <" + edgeElement.getName()
						+ "> unknown");
			}
		}
		connect(getPortFromNode(sourceId), getPortFromNode(targetId), type);
	}

	private Port getPortFromNode(String id) {
		WorkflowNode node = getNode(id);
		if (node != null) {
			DefaultPort port = new DefaultPort();
			((DefaultGraphCell) node).add(port);
			return port;
		}
		return null;
	}

	private WorkflowNode getNode(String id) {
		for (Object cell : getRoots()) {
			if (cell instanceof WorkflowNode) {
				WorkflowNode node = (WorkflowNode) cell;
				if (node.getIdName().equals(id)) {
					return node;
				}
			}
		}
		return null;
	}

	// Insert a new Edge between source and target
	public void connect(Port sourcePort, Port targetPort, String label) {
		if (sourcePort != null && targetPort != null) {
			// Construct Edge with no label
			Edge edge = new Edge(generateNewEdgeID());
			if (getModel().acceptsSource(edge, sourcePort)
					&& getModel().acceptsTarget(edge, targetPort)) {
				WorkflowNode sourceNode = (WorkflowNode) ((DefaultPort) sourcePort)
						.getRoot();
				WorkflowNode targetNode = (WorkflowNode) ((DefaultPort) targetPort)
						.getRoot();
				if (sourceNode.isConnectableWith(targetNode)) {
					if (!edgeExists(sourceNode, targetNode)) {
						edge.setSourceAndTargetTypes(sourceNode.getType(),
								targetNode.getType());
						edge.setSourceNode(sourceNode);
						edge.setTargetNode(targetNode);
						if (label != null && edge.isLabelEditable()) {
							edge.setType(label);
						} else {
							String label2 = Edge.LABELS_MATRIX[sourceNode
									.getType()][targetNode.getType()];
							if (edge.isFromToolToDataset()) {
								/*
								 * Connecting a tool with a dataset node: if
								 * this is a multiple connection then change the
								 * edge index
								 */
								int numOutputEdges = getOutgoingEdges(
										sourceNode).size();
								if (numOutputEdges > 0) {
									if (numOutputEdges == 1) {
										getFirstOutgoingEdge(sourceNode)
												.setIndex(1);
									}
									// TODO ### differenciate among datasets,
									// trainingsets and testsets
									edge.setIndex(numOutputEdges + 1);
								}
							}
							edge.setType(label2);
						}
						// Create a Map thath holds the attributes for the edge
						edge.getAttributes().applyMap(createEdgeAttributes());
						// Insert the Edge and its Attributes
						getGraphLayoutCache().insertEdge(edge, sourcePort,
								targetPort);
					} else {
						JOptionPane
								.showMessageDialog(
										view.getFrame(),
										" A connection from the two nodes already exists.",
										"Connection error",
										JOptionPane.ERROR_MESSAGE);
					}

				} else {
					JOptionPane.showMessageDialog(view.getFrame(),
							"Connection from a " + sourceNode.getName()
									+ " to a " + targetNode.getName()
									+ " is not allowed.", "Connection error",
							JOptionPane.ERROR_MESSAGE);
				}
			}
		} else {
			System.err.println("Source and/or target node don't exist");
		}
	}

	private boolean edgeExists(WorkflowNode source, WorkflowNode target) {
		for (Object cell : getRoots()) {
			if (cell instanceof Edge) {
				WorkflowNode edgeSource = (WorkflowNode) ((DefaultPort) ((Edge) cell)
						.getSource()).getRoot();
				WorkflowNode edgeTarget = (WorkflowNode) ((DefaultPort) ((Edge) cell)
						.getTarget()).getRoot();
				if ((edgeSource.equals(source) && edgeTarget.equals(target))
						|| (edgeSource.equals(target) && edgeTarget
								.equals(source))) {
					return true;
				}
			}
		}
		return false;
	}

	// Hook for subclassers
	@SuppressWarnings("unchecked")
	private Map<?, ?> createEdgeAttributes() {
		Map<?, ?> map = new Hashtable();
		// Add a Line End Attribute
		GraphConstants.setLineEnd(map, GraphConstants.ARROW_SIMPLE);
		// Add a label along edge attribute
		GraphConstants.setLabelAlongEdge(map, true);
		return map;
	}

	public void saveToFile() {
		saveToFile(fileFullPath);
	}

	/*
	 * Does a breadth-first search (BFS) graph traversal and update the edges
	 * ids progressively
	 */
	private void updateEdgeIds() {
		Queue<WorkflowNode> queue = new LinkedList<WorkflowNode>();
		Set<WorkflowNode> done = new HashSet<WorkflowNode>();
		for (WorkflowNode root : getRootNodes()) {
			queue.add(root);
			done.add(root);
		}
		int id = 0;
		while (!queue.isEmpty()) {
			WorkflowNode node = queue.remove();
			for (Edge e : getOutgoingEdges(node)) {
				e.setId(id++);
				WorkflowNode targetNode = e.getTargetNode();
				if (!done.contains(targetNode)) {
					done.add(targetNode);
					queue.add(targetNode);
				}
			}
		}
	}

	public void saveToFile(String newFileName) {
		fileFullPath = newFileName;

		Document doc = new Document();
		Element graphml = new Element("graphml");
		graphml.setNamespace(NAMESPACE);
		Element graph = new Element("graph");
		graph.setAttribute("id", "G");
		graph.setAttribute("edgedefault", "directed");

		/*
		 * Necessary for the EPMS as it requires the edges ids to be
		 * progressively numbered from the root nodes to the leaf nodes
		 */
		updateEdgeIds();

		ArrayList<GraphElement> graphElements = new ArrayList<GraphElement>(
				getRoots().length);

		for (Object cell : getRoots()) {
			if (cell instanceof WorkflowNode) {
				graphElements.add(((WorkflowNode) cell).getElement());
			} else if (cell instanceof Edge) {
				graphElements.add(((Edge) cell).getElement());
			}
		}

		// Necessary for the EPMS
		Collections.sort(graphElements);

		for (Element element : graphElements) {
			graph.addContent(element);
		}

		graphml.addContent(graph);
		doc.addContent(graphml);

		XMLOutputter outp = new XMLOutputter();
		outp.setFormat(Format.getPrettyFormat());

		try {
			outp.output(doc, new FileOutputStream(newFileName));
		} catch (FileNotFoundException e) {
			JOptionPane.showMessageDialog(view.getFrame(), e,
					"FileNotFoundException", JOptionPane.ERROR_MESSAGE);
		} catch (IOException e) {
			JOptionPane.showMessageDialog(view.getFrame(), e, "IOException",
					JOptionPane.ERROR_MESSAGE);
		}
		changed = false;
	}

	/*
	 * Returns a list of root nodes, that is the nodes which don't have incoming
	 * edges.
	 */
	private LinkedList<WorkflowNode> getRootNodes() {
		LinkedList<WorkflowNode> rootNodes = new LinkedList<WorkflowNode>();
		for (WorkflowNode node : getNodes()) {
			if (node.isRoot()) {
				rootNodes.add(node);
			}
		}
		return rootNodes;
	}

	/*
	 * Returns a list of leaf nodes, that is the nodes which don't have outgoing
	 * edges.
	 */
	private LinkedList<WorkflowNode> getLeafNodes() {
		LinkedList<WorkflowNode> leafNodes = new LinkedList<WorkflowNode>();
		for (WorkflowNode node : getNodes()) {
			if (node.isLeaf()) {
				leafNodes.add(node);
			}
		}
		return leafNodes;
	}

	public String getFileFullPath() {
		return fileFullPath;
	}

	public String getFileName() {
		String fileNameWithExtension = fileFullPath.substring(fileFullPath
				.lastIndexOf(File.separator) + 1);
		System.out.println(fileNameWithExtension);
		if (fileNameWithExtension.lastIndexOf(".") > 0) {
			return fileNameWithExtension.substring(0, fileNameWithExtension
					.lastIndexOf("."));
		}
		return fileNameWithExtension;
	}

	public void enableSelection(boolean b) {
		if (b) {
			setMarqueeHandler(basicMarqueeHandler);
		} else {
			setMarqueeHandler(customMarqueeHandler);
		}
	}

	public boolean isEmpty() {
		return getRoots().length == 0;
	}

	public String saveToTempFile() {
		String originalFile = fileFullPath;
		boolean originalChanged = changed;
		saveToFile(TEMP_FILE_NAME);
		fileFullPath = originalFile;
		changed = originalChanged;
		return TEMP_FILE_NAME;
	}

	public void update(String id, String status) {
		if (id.startsWith("n")) {
			try {
				Integer.parseInt(id.substring(1));
				WorkflowNode node = getNode(id);
				if (node != null) {
					node.setStatus(status);
					if (running) {
						checkTermination(node);
					}
					// graphLayoutCache.showCells(new Object[] { node }, false);
					graphLayoutCache.reload();
					// revalidate();
					refresh();
				} else {
					Utils
							.print("Status update error: No node found with the id = "
									+ id);
				}
			} catch (NumberFormatException e) {
				Utils.print("Status update error: the id \"" + id
						+ "\" is not a number");
			} catch (JDOMException e) {
				Utils.print(e.getMessage());
			}
		}
	}

	private void checkTermination(WorkflowNode node) {
		int status = node.getStatus();
		if (status == NOT_FOUND || status == FAILED) {
			termination(false);
		} else if (node.isLeaf()) {
			if (status == DONE) {
				for (WorkflowNode n : getLeafNodes()) {
					if (n.getStatus() != DONE) {
						return;
					}
				}
				termination(true);
			}
		}
	}

	private void termination(boolean showMessage) {
		running = false;
		stopWatch.stop();
		String message = "The workflow \""
				+ getFileName()
				+ "\" has successfully terminated in "
				+ DurationFormatUtils.formatDurationWords(stopWatch.getTime(),
						true, false);
		Utils.print(message);
		if (showMessage) {
			new Thread(message) {
				@Override
				public void run() {
					JOptionPane.showMessageDialog(view.getFrame(), getName(),
							"Success", JOptionPane.INFORMATION_MESSAGE);
				}
			}.start();
		}
	}

	public void resetNodesStatuses() {
		for (Object cell : getRoots()) {
			if (cell instanceof WorkflowNode) {
				WorkflowNode node = (WorkflowNode) cell;
				node.resetStatus();
			}
		}
		getGraphLayoutCache().showCells(getRoots(), false);
		getGraphLayoutCache().reload();
	}

	public String getNodeType(String id) {
		if (id.startsWith("n")) {
			try {
				Integer.parseInt(id.substring(1));
				return getNode(id).getName();
			} catch (NumberFormatException e) {
				kgrid.client.gui.Utils.print("ERROR: the id \"" + id
						+ "\" is not a number");
			}
		}
		return "";
	}

	/**
	 * Checks whether all root datasets (i.e. those not having incoming edges)
	 * and all tool nodes have at least one property set. Returns a string
	 * containing the concatenation of the error messages of the nodes without
	 * properties set (called empty nodes). If all nodes contain at least one
	 * property it returns an empty string.
	 */
	public String getEmptyNodesErrorMessages() {
		String errorMessages = "";
		for (WorkflowNode node : getNodes()) {
			if (node instanceof DatasetNode && node.isRoot()
					&& ((DatasetNode) node).isEmpty()) {
				errorMessages += "The dataset node "
						+ ((DatasetNode) node).getIdName()
						+ " must contain at least one property.\n";

			} else if (node instanceof ToolNode && ((ToolNode) node).isEmpty()) {
				errorMessages += "The tool node "
						+ ((ToolNode) node).getIdName()
						+ " must contain at least one property.\n";
			}
		}
		return errorMessages;
	}

	/*
	 * It is invoked when an edge or node is deleted: it updates the edges
	 * labels with an index which are related with any of the workflow elements
	 * contained in the 'cells' list.
	 */
	private void updateOtherEdges(Object[] cells) {
		for (Object cell : cells) {
			if (cell instanceof Edge && ((Edge) cell).isFromToolToDataset()) {
				ArrayList<Edge> edges = getOutgoingEdges(((Edge) cell)
						.getSourceNode());
				if (edges.size() == 2) {
					for (Edge edge : edges) {
						if (!((Edge) cell).equals(edge)) {
							edge.setIndex(0);
							break;
						}
					}
				} else if (edges.size() > 2) {
					int i = 1;
					for (Edge edge : edges) {
						if (!((Edge) cell).equals(edge)) {
							edge.setIndex(i++);
						}
					}
				}
			}
		}
	}

	public void removeSelectedCells() {
		// cells = graph.getDescendants(cells);
		Object[] cells = getSelectionCells();
		updateOtherEdges(cells);
		Object[] children = DefaultGraphModel.getEdges(getModel(), cells)
				.toArray();
		updateOtherEdges(children);
		getModel().remove(getDescendants(cells));
		getModel().remove(children);
		getModel().remove(cells);
		getGraphLayoutCache().reload();
	}
}
