/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.client.workflow;

import java.awt.geom.Rectangle2D;
import java.util.LinkedHashMap;

import org.jdesktop.application.ResourceMap;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultPort;
import org.jgraph.graph.GraphConstants;

/**
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public abstract class WorkflowNode extends DefaultGraphCell implements Consts {
	private static final long serialVersionUID = 1L;
	private ResourceMap resourceMap;
	private Workflow workflow;

	private String name;
	protected int id;
	protected int status = NOINFO;
	protected LinkedHashMap<String, Object> properties = new LinkedHashMap<String, Object>();

	public WorkflowNode(int id, String name, int x, int y,
			ResourceMap resourceMap, Workflow w) {
		this.resourceMap = resourceMap;
		this.name = name;
		this.id = id;
		this.workflow = w;
		add(new DefaultPort());
		// GraphConstants.setLabelEnabled(getAttributes(), true);
		updateIcon();
		GraphConstants.setBounds(getAttributes(), new Rectangle2D.Double(x
				- BOUND_SIZE / 2, y - BOUND_SIZE / 2, BOUND_SIZE, BOUND_SIZE));
		GraphConstants.setInset(getAttributes(), 4);
		GraphConstants.setAutoSize(getAttributes(), true);
	}

	public String getName() {
		return name;
	}

	public String getIdName() {
		return "n" + id;
	}

	public int getIdNumber() {
		return id;
	}

	public void setId(String idName) {
		id = Integer.valueOf(idName.substring(1));
	}

	public String getStatusString() {
		switch (status) {
		case NOINFO:
			return "No information provided";

		case LOCATION_SET:
			return "Location set";

		case DONE:
			return "Task completed successfully";

		case FAILED:
			return "Execution failed";

		case MATCHED:
			return "Matching resource found...";

		case RUNNING:
			return "Running...";

		case NOT_FOUND:
			return "Resource not found in the Knowledge Grid";

		case KDSURL_SET:
			return "KDS URL set";

		case SEARCHP_SET:
			return "Search parameters set";

		default:
			break;
		}
		return null;
	}

	public int getStatus() {
		return status;
	}

	public abstract int getType();

	public void setStatus(int status) {
		this.status = status;
		updateIcon();
	}

	protected abstract void setKDSURL(String s);

	public void setStatus(String status) throws JDOMException {
		// TODO KGClient: save status string-int mapping in a hashmap
		if (status == null || status.equals("noinfo")) {
			setStatus(NOINFO);
		} else if (status.equals("done") || status.equals("DONE")) {
			setStatus(DONE);
		} else if (status.equals("unavailable")) {
			setStatus(UNAVAILABLE);
		} else if (status.equals("error") || status.startsWith("FAILED")
				|| status.startsWith("ERROR")) {
			if (FAILED > this.status) {
				setStatus(FAILED);
			}
		} else if (status.equals("ready")) {
			setStatus(LOCATION_SET);
		} else if (status.startsWith("available resources")) {
			if (NOT_FOUND > this.status) {
				if (status.substring(status.indexOf("=") + 1).isEmpty()) {
					setStatus(NOT_FOUND);
				}
			}
		} else if (status.startsWith("kds")) {
			setKDSURL(status.substring(0, status.indexOf(" ")));
			if (MATCHED > this.status) {
				setStatus(MATCHED);
			}
		} else if (status.equals("running") || status.equals("ACTIVE")) {
			if (RUNNING > this.status) {
				setStatus(RUNNING);
			}
		} else if (status.equals("DESTROYED") || status.equals("CREATED")) {
			// do nothing
		} else {
			throw new JDOMException("Status \"" + status + "\" unknown");
		}
	}

	public LinkedHashMap<String, Object> getProperties() {
		return properties;
	}

	public boolean isConnectableWith(WorkflowNode node) {
		return CONNECTION_ALLOWED[getType()][node.getType()];
	}

	protected void addPosition(Element node) {
		Rectangle2D rect = GraphConstants.getBounds(getAttributes());

		Element posXE = new Element("data");
		posXE.setAttribute("key", "position_X");
		posXE.addContent(Integer.toString((int) rect.getCenterX()));
		node.addContent(posXE);

		Element posYE = new Element("data");
		posYE.setAttribute("key", "position_Y");
		posYE.addContent(Integer.toString((int) rect.getCenterY()));
		node.addContent(posYE);
	}

	/**
	 * Set the node property named {@code property} with the value {@code value}
	 * . In case of properties with the same name then the {@code row} table
	 * index is used.
	 * 
	 * @param row
	 * @param property
	 * @param value
	 */
	public abstract void setProperty(int row, String property, String value);

	public abstract void setProperties(Object e) throws JDOMException;

	public abstract GraphElement getElement();

	public abstract boolean isEmpty();

	public void updateIcon() {
		GraphConstants.setIcon(getAttributes(),
				resourceMap.getIcon(name + status + ".icon"));
	}

	public abstract String getKDSURL();

	public abstract void resetStatus();

	public abstract void clearSearchParameters();

	/**
	 * Checks whether the node 'node' is a root node (i.e. it doesn't have any
	 * incoming edge)
	 */
	public boolean isRoot() {
		for (Edge e : workflow.getEdges()) {
			if (e.getTargetNode().equals(this)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Checks whether the node 'node' is a leaf node (i.e. it doesn't have any
	 * outgoing edge)
	 */
	public boolean isLeaf() {
		for (Edge e : workflow.getEdges()) {
			if (e.getSourceNode().equals(this)) {
				return false;
			}
		}
		return true;
	}
}